/* 
 * nre.c --
 *
 *	This file contains the commands for the nre package.
 *
 * Copyright (c) 1997 Darrel Schneider
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */
#include <tcl.h>
#include <string.h>
#include "nre.h"

static void	DupRegExpInternalRep _ANSI_ARGS_((Tcl_Obj *srcPtr,
						  Tcl_Obj *copyPtr));
static void	FreeRegExpInternalRep _ANSI_ARGS_((Tcl_Obj *rePtr));
static int	SetRegExpFromAny _ANSI_ARGS_((Tcl_Interp *interp,
					      Tcl_Obj *objPtr));
static void	UpdateStringOfRegExp _ANSI_ARGS_((Tcl_Obj *rePtr));

static int      okToTrash _ANSI_ARGS_((long *ssPtr,
				       long *ssEnd,
				       Tcl_Obj* patternObj,
				       Tcl_RegExpExecResult *reResPtr,
				       char *endTrashed));
static int      parseSubSpec _ANSI_ARGS_((char* string,
					  long strSize,
					  long** stackBuffPtr,
					  int  buffSize));

extern
#if defined(_WIN32) || defined(__WIN32__)
__declspec(dllexport)
#endif
int Nre_Init _ANSI_ARGS_((Tcl_Interp* interp));

/*
 * The structure below defines the list Tcl object type by means of
 * procedures that can be invoked by generic object code.
 */

static Tcl_ObjType tclRegExpType = {
    "regexp",				/* name */
    FreeRegExpInternalRep,		/* freeIntRepProc */
    DupRegExpInternalRep,	        /* dupIntRepProc */
    UpdateStringOfRegExp,		/* updateStringProc */
    SetRegExpFromAny			/* setFromAnyProc */
};

#define NRE_CACHE_SIZE 7

/* nrecache should really be part of the interp state in tclInt.h
 * Right now the cache is only used for the top-level interpreter.
 */
static Tcl_Obj* nrecache[NRE_CACHE_SIZE] = {NULL};
static Tcl_Obj* lastFoundInCache = NULL;

static void
addToNreCache(interp, objPtr)
    Tcl_Interp *interp;		/* Used for error reporting if not NULL. */
    Tcl_Obj *objPtr;		/* The object to consider adding. */
{
  int i;
  /* PRECONDITION: objPtr->typePtr == &tclRegExpType */

  if (Tcl_IsShared(objPtr)) {
    /* don't bother adding it since the user has probably already
     * cached it in a variable.
     */
    return;
  }
  if (Tcl_GetMaster(interp) != NULL) {
    /* This cache is only contains objects that belong to the top level
     * interpreter. That way we won't get in trouble by caching on object
     * of an interpreter that get blown away.
     * Since 'interp' had a master we know it was not the top level.
     */
    return;
  }
  
  if (nrecache[NRE_CACHE_SIZE-1] != NULL) {
    /* cache is full. So release the last element. */
    Tcl_DecrRefCount(nrecache[NRE_CACHE_SIZE-1]);
    nrecache[NRE_CACHE_SIZE-1] = NULL;
  }
  for (i = NRE_CACHE_SIZE-1; i > 0; i--) {
    nrecache[i] = nrecache[i-1];
  }
  nrecache[0] = objPtr;
  /* the cache will have a reference so increment it */
  Tcl_IncrRefCount(objPtr);
}

static int
findInNreCache(interp, objPtr, str, length)
    Tcl_Interp *interp;		/* Used for error reporting if not NULL. */
    Tcl_Obj *objPtr;		/* The object to find. */
    char* str;                  /* the object's string */
    int length;                 /* the object's string's length */
{
  int j;
  int idx = 0;
  Tcl_Obj* cacheObjPtr;

  while ((cacheObjPtr = nrecache[idx]) != NULL) {
    if (cacheObjPtr->typePtr != &tclRegExpType) {
      /* This guy is no longer a regexp object so remove it from the cache */
      Tcl_DecrRefCount(cacheObjPtr);
      for (j = idx; j < NRE_CACHE_SIZE-1; j++) {
	nrecache[j] = nrecache[j+1];
      }
      nrecache[NRE_CACHE_SIZE-1] = NULL;
      /* note we do not increment idx since everything was moved down */
    } else {
      /* Check to see if the cacheObj is the same as the current guy */
      int cacheStrLen;
      char* cacheStr = Tcl_GetStringFromObj(cacheObjPtr, &cacheStrLen);
      if (cacheStrLen == length) {
	if (0 == memcmp(str, cacheStr, length)) {
	  /* move this guy to slot 0 since he was used */
	  if (idx != 0) {
	    for (j = idx; j > 0; j--) {
	      nrecache[j] = nrecache[j-1];
	    }
	    nrecache[0] = cacheObjPtr;
	  }
	  DupRegExpInternalRep(cacheObjPtr, objPtr);
	  lastFoundInCache = objPtr;
	  return 1;
	}
      }
      idx++;
      if (idx == NRE_CACHE_SIZE) {
	return 0;
      }
    }
  }
  return 0;
}
/*
 *----------------------------------------------------------------------
 *
 * FreeRegExpInternalRep --
 *
 *	Deallocate the storage associated with a regexp object's internal
 *	representation.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Frees rePtr's List* internal representation and sets rePtr's
 *	internalRep.otherValuePtr to NULL.
 *
 *----------------------------------------------------------------------
 */

static void
FreeRegExpInternalRep(rePtr)
    Tcl_Obj *rePtr;		/* RegExp object with internal rep to free. */
{
    ckfree((char *) rePtr->internalRep.otherValuePtr);
    rePtr->internalRep.otherValuePtr = NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * DupRegExpInternalRep --
 *
 *	Initialize the internal representation of a regexp Tcl_Obj to a
 *	copy of the internal representation of an existing regexp object. 
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	"srcPtr"s regexp internal rep pointer should not be NULL and we assume
 *	it is not NULL. We set "copyPtr"s internal rep to a pointer to a
 *	newly allocated Tcl_RegExp structure that, in turn, is set to contain
 *      a copy of srcPtr.
 *
 *----------------------------------------------------------------------
 */

static void
DupRegExpInternalRep(srcPtr, copyPtr)
    Tcl_Obj *srcPtr;		/* Object with internal rep to copy. */
    Tcl_Obj *copyPtr;		/* Object with internal rep to set. */
{
    regexp *srcRepPtr = (regexp*) srcPtr->internalRep.otherValuePtr;
    regexp *cpyRepPtr;
    int psize = srcRepPtr->regexpSize;

    cpyRepPtr = (regexp*) ckalloc(psize);
    memcpy(cpyRepPtr, srcRepPtr, psize);

    /* fixup the fields that are pointers */
    if (cpyRepPtr->regstartnode != NULL) {
      cpyRepPtr->regstartnode -= (unsigned long)srcRepPtr;
      cpyRepPtr->regstartnode += (unsigned long)cpyRepPtr;
    }
    if (cpyRepPtr->regmust != NULL) {
      cpyRepPtr->regmust -= (unsigned long)srcRepPtr;
      cpyRepPtr->regmust += (unsigned long)cpyRepPtr;
    }
    if (cpyRepPtr->regclasses != NULL) {
      cpyRepPtr->regclasses -= (unsigned long)srcRepPtr;
      cpyRepPtr->regclasses += (unsigned long)cpyRepPtr;
    }
    
    copyPtr->internalRep.otherValuePtr = (VOID *) cpyRepPtr;
    copyPtr->typePtr = &tclRegExpType;
}

/*
 *----------------------------------------------------------------------
 *
 * SetRegExpFromAny --
 *
 *	Attempt to generate a regexp internal form for the Tcl object
 *	"objPtr".
 *
 * Results:
 *	The return value is TCL_OK or TCL_ERROR. If an error occurs during
 *	conversion, an error message is left in the interpreter's result
 *	unless "interp" is NULL.
 *
 * Side effects:
 *	If no error occurs, a regexp is stored as "objPtr"s internal
 *	representation.
 *
 *----------------------------------------------------------------------
 */

static int
SetRegExpFromAny(interp, objPtr)
    Tcl_Interp *interp;		/* Used for error reporting if not NULL. */
    Tcl_Obj *objPtr;		/* The object to convert. */
{
    regexp *rePtr;
    Tcl_ObjType *oldTypePtr = objPtr->typePtr;
    char *string;
    int length;

    if (oldTypePtr == &tclRegExpType) {
      /* Already is a regexp object */
      return TCL_OK;
    }

    /*
     * Get the string representation. Make it up-to-date if necessary.
     */

    string = Tcl_GetStringFromObj(objPtr, &length);

    /*
     * Free the old internalRep before setting the new one.
     * Note we do this after Tcl_GetStringFromObj.
     */
    if ((oldTypePtr != NULL) && (oldTypePtr->freeIntRepProc != NULL)) {
	oldTypePtr->freeIntRepProc(objPtr);
        objPtr->typePtr = NULL;
    }

    if (!findInNreCache(interp, objPtr, string, length)) {
      TclNewRegError((char *) NULL);
      rePtr = TclNewRegCompSz(string, length, 0);
      if (TclNewGetRegError() != NULL) {
	Tcl_AppendResult(interp,
			 "couldn't compile regular expression pattern: ",
			 TclNewGetRegError(), (char *) NULL);
	return TCL_ERROR;
      }

      objPtr->internalRep.otherValuePtr = (VOID *) rePtr;
      objPtr->typePtr = &tclRegExpType;
      addToNreCache(interp, objPtr);
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * UpdateStringOfRegExp --
 *
 *	Update the string representation for a regexp object.
 *	Note: regexp object's are invariant so whatever string
 *      they had when created or dupped is ok. So this function is a noop.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static void
UpdateStringOfRegExp(rePtr)
    Tcl_Obj *rePtr;
{
    if (rePtr->bytes == NULL) {
#if 0 /* use this code if nre is moved into the core */
	rePtr->bytes = tclEmptyStringRep;
#else
	rePtr->bytes = (char *) ckalloc((unsigned) 1);
	rePtr->bytes[0] = '\0';
#endif
	rePtr->length = 0;
    }
    return;
  /* do nothing */
}

static Tcl_Obj*
Tcl_RegExpCompileObj(interp, pattern, noCase)
    Tcl_Interp *interp;			/* For use in error reporting. */
    Tcl_Obj *pattern;			/* String for which to produce
					 * compiled regular expression. */
    int noCase;                         /* if 1 case insensitive */
{
  /*
   * If noCase is 1 then the user explicity asked for case insensitive.
   * Otherwise it is up to the pattern's contents (i.e. (?i) switch).
   */
  if (pattern->typePtr == &tclRegExpType) {
    int allDone = 1;
    if (noCase) {
      /* Need to check pattern's state and see if it is noCase */
      regexp *rePtr = (regexp*)pattern->internalRep.otherValuePtr;
      if (!(rePtr->regflags & REGEXP_NOCASE)) {
	/* Need to recompile
	 * NYI: should check if it is shared and if so create a new object!
	 */
	allDone = 0;
      }
    }
    if (allDone) {
      return pattern;
    }
  }
  if (noCase) {
    /* inline version of setRegExpFromAny.
     * Need this hack so we can specify noCase switch.
     */
    regexp *rePtr;
    Tcl_ObjType *oldTypePtr = pattern->typePtr;
    char *string;
    int length;

    string = Tcl_GetStringFromObj(pattern, &length);

    /*
     * Free the old internalRep before setting the new one.
     * Note we do this after Tcl_GetStringFromObj.
     */
    if ((oldTypePtr != NULL) && (oldTypePtr->freeIntRepProc != NULL)) {
	oldTypePtr->freeIntRepProc(pattern);
        pattern->typePtr = NULL;
    }

    TclNewRegError((char *) NULL);
    rePtr = TclNewRegCompSz(string, length, noCase);
    if (TclNewGetRegError() != NULL) {
	Tcl_AppendResult(interp,
	    "couldn't compile regular expression pattern: ",
	    TclNewGetRegError(), (char *) NULL);
	return NULL;
    }

    pattern->internalRep.otherValuePtr = (VOID *) rePtr;
    pattern->typePtr = &tclRegExpType;
    return pattern;
  } else {
    if (Tcl_ConvertToType(interp, pattern, &tclRegExpType) == TCL_OK) {
      return pattern;
    } else {
      return NULL;
    }
  }
}

static int
Tcl_RegExpExecObj(interp, reObj, string, start, stringLength, resptr,
		  couldHavePtr)
    Tcl_Interp *interp;		/* Interpreter to use for error reporting. */
    Tcl_Obj* reObj;		/* Compiled regular expression;  must have
				 * been returned by previous call to
				 * Tcl_RegExpCompileObj. */
    char *string;		/* String against which to match re. */
    char *start;		/* If string is part of a larger string,
				 * this identifies beginning of larger
				 * string, so that "^" won't match. */
    int stringLength;
    Tcl_RegExpExecResult *resptr;
    int* couldHavePtr; /* if not NULL then set to 1 if the reObj could have
			* matched if more input was provided.
			*/
{
    int match;
    regexp *regexpPtr = (regexp*)reObj->internalRep.otherValuePtr;

    if (reObj->typePtr != &tclRegExpType) {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, "expect a regular expression", (char *) NULL);
	return -1;
    }

    TclNewRegError((char *) NULL);
    match = TclNewRegExecSz(regexpPtr, string, start, stringLength, resptr,
			    couldHavePtr);
    if (TclNewGetRegError() != NULL) {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, "error while matching regular expression: ",
		TclNewGetRegError(), (char *) NULL);
	return -1;
    }
    return match;
}

static void
Tcl_NewRegExpRange(re, resptr, index, startPtr, endPtr)
    Tcl_RegExp re;		/* Compiled regular expression that has
				 * been passed to Tcl_RegExpExec. */
    Tcl_RegExpExecResult *resptr;           /* Result from Tcl_RegExpExec */
    int index;			/* 0 means give the range of the entire
				 * match, > 0 means give the range of
				 * a matching subrange.  Must be no greater
				 * than NSUBEXP. */
    char **startPtr;		/* Store address of first character in
				 * (sub-) range here. */
    char **endPtr;		/* Store address of character just after last
				 * in (sub-) range here. */
{
    regexp *regexpPtr = (regexp *) re;

    if (index >= regexpPtr->subexpCount) {
	*startPtr = *endPtr = NULL;
    } else {
	*startPtr = resptr->startp[index];
	*endPtr = resptr->endp[index];
    }
}
#define REGEXP_RANGE0_START(x) (x.startp[0])
#define REGEXP_RANGE0_END(x) (x.endp[0])

/* Object form of Tcl_RegExpRange.
 */
static void
Tcl_RegExpRangeObj(reObj, resptr, index, startPtr, endPtr)
    Tcl_Obj* reObj;		/* Compiled regular expression that has
				 * been passed to Tcl_RegExpExec. */
    Tcl_RegExpExecResult *resptr;           /* Result from Tcl_RegExpExec */
    int index;			/* 0 means give the range of the entire
				 * match, > 0 means give the range of
				 * a matching subrange.  Must be no greater
				 * than NSUBEXP. */
    char **startPtr;		/* Store address of first character in
				 * (sub-) range here. */
    char **endPtr;		/* Store address of character just after last
				 * in (sub-) range here. */
{
  Tcl_NewRegExpRange((Tcl_RegExp)reObj->internalRep.otherValuePtr,
		     resptr, index, startPtr, endPtr);
}

static int
Tcl_RegExpMatchObj(interp, string, stringLength, reObj)
    Tcl_Interp *interp;		/* Used for error reporting. */
    char *string;		/* String. */
    int stringLength;
    Tcl_Obj* reObj;
{
    Tcl_RegExpExecResult reres;
    if (Tcl_ConvertToType(interp, reObj, &tclRegExpType) != TCL_OK) {
      return -1;
    }
    return Tcl_RegExpExecObj(interp, reObj, string, string, stringLength,
			     &reres, NULL);
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_NewLsearchObjCmd --
 *
 *	This procedure is invoked to process the "nrelsearch" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

static int
Tcl_NewLsearchObjCmd(clientData, interp, objc, objv)
    ClientData clientData;	/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj *CONST objv[];	/* Argument values. */
{
#define EXACT	0
#define GLOB	1
#define REGEXP	2
    char *bytes, *patternBytes;
    Tcl_Obj* patternObj;
    int i, match, mode, index, result, listLen, length, elemLen;
    Tcl_Obj **elemPtrs;
    static char *switches[] =
	    {"-exact", "-glob", "-regexp", (char *) NULL};

    mode = GLOB;
    if (objc == 4) {
	if (Tcl_GetIndexFromObj(interp, objv[1], switches,
		"search mode", 0, &mode) != TCL_OK) {
	    return TCL_ERROR;
	}
    } else if (objc != 3) {
	Tcl_WrongNumArgs(interp, 1, objv, "?mode? list pattern");
	return TCL_ERROR;
    }

    /*
     * Make sure the list argument is a list object and get its length and
     * a pointer to its array of element pointers.
     */

    result = Tcl_ListObjGetElements(interp, objv[objc-2], &listLen, &elemPtrs);
    if (result != TCL_OK) {
	return result;
    }

    patternObj = objv[objc-1];
    if (mode != REGEXP) {
      patternBytes = Tcl_GetStringFromObj(patternObj, &length);
    }
 
    index = -1;
    for (i = 0; i < listLen; i++) {
	match = 0;
	bytes = Tcl_GetStringFromObj(elemPtrs[i], &elemLen);
	switch (mode) {
	    case EXACT:
		if (length == elemLen) {
		    match = (memcmp(bytes, patternBytes,
			    (size_t) length) == 0);
		}
		break;
	    case GLOB:
		/*
		 * WARNING: will not work with data containing NULLs.
		 */
		match = Tcl_StringMatch(bytes, patternBytes);
		break;
	    case REGEXP:
		match = Tcl_RegExpMatchObj(interp, bytes, elemLen, patternObj);
		if (match < 0) {
		    return TCL_ERROR;
		}
		break;
	}
	if (match) {
	    index = i;
	    break;
	}
    }

    Tcl_SetIntObj(Tcl_GetObjResult(interp), index);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_NewRegexpObjCmd --
 *
 *	This procedure is invoked to process the "nrematch" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

typedef struct {
  int linenum;
  int offset;
} textidxSType;

typedef struct {
  int linenum;
  char* lptr; /* The first character of the linenum line */
  char* nextlptr; /* The first character of the next line */
  char* stringEnd;
  Tcl_Obj* patternObj;
} tiStateSType;

static void calcTi _ANSI_ARGS_((textidxSType *tiPtr,
				char *p, tiStateSType *tiState));
static void calcAllVarTis _ANSI_ARGS_((textidxSType *startTiPtr,
				       textidxSType *endTiPtr,
				       tiStateSType *tiStatePtr,
				       Tcl_RegExpExecResult *reResPtr,
				       int count));
static Tcl_Obj* tiToObj _ANSI_ARGS_((textidxSType* tiPtr));


static void calcTi (tiPtr, p, tiState)
  textidxSType *tiPtr;
  char *p;
  tiStateSType *tiState;
{
  int linenum = tiState->linenum;
  char *lptr = tiState->lptr;
  char *nextlptr = tiState->nextlptr;
  if (p >= nextlptr) {
    do {
      int count = tiState->stringEnd - nextlptr;
      lptr = nextlptr;
      nextlptr = (char*)memchr(nextlptr, '\n', count);
      if (!nextlptr) {
	/* at the end */
	nextlptr = tiState->stringEnd + 1;
      } else {
	nextlptr++;
      }
      linenum++;
    } while (p >= nextlptr);
    tiState->linenum = linenum;
    tiState->lptr = lptr;
    tiState->nextlptr = nextlptr;
  }
#if 0
  if (p < lptr) {
    /* This should never happen.
     * If it can then this function will need to scan backwards.
     */
    fprintf(stderr, "p=%p < lptr=%p\n", p, lptr);
    exit(1);
  }
#endif
  /* At this point I know p >= lptr and < nextlptr */
  tiPtr->linenum = linenum;
  tiPtr->offset = p - lptr;
}

static void calcAllVarTis(startTiPtr, endTiPtr, tiStatePtr, reResPtr, count)
     textidxSType *startTiPtr;
     textidxSType *endTiPtr;
     tiStateSType *tiStatePtr;
     Tcl_RegExpExecResult *reResPtr;
     int count;
{
  /* This function may convert more than is actually needed */
  struct tiSortRec {
    struct tiSortRec *next;
    char* ptr;
    textidxSType* tiPtr;
    struct tiSortRec *prev;
  };
  int i;
  char* first;
  char* last;
  struct tiSortRec* sortRecPtr;
  struct tiSortRec* lastAddedPtr;
  struct tiSortRec sortTable[NEWNSUBEXP*2];
  int stSize = 2;
  Tcl_RegExpRangeObj(tiStatePtr->patternObj, reResPtr, 0, &first, &last);
  sortTable[0].next = &sortTable[1];
  sortTable[0].ptr = first;
  sortTable[0].tiPtr = &startTiPtr[0];
  sortTable[0].prev = NULL;
  sortTable[1].next = NULL;
  sortTable[1].ptr = last;
  sortTable[1].tiPtr = &endTiPtr[0];
  sortTable[1].prev = &sortTable[0];
  sortRecPtr = &sortTable[1];
  for (i=1; i < count; i++) {
    Tcl_RegExpRangeObj(tiStatePtr->patternObj, reResPtr, i, &first, &last);
    if (first != NULL) {
      lastAddedPtr = sortRecPtr;
      sortRecPtr = &sortTable[stSize];
      stSize++;
      sortRecPtr->ptr = first;
      sortRecPtr->tiPtr = &startTiPtr[i];
      if (first > lastAddedPtr->ptr) {
	do {
	  lastAddedPtr = lastAddedPtr->next;
	} while (first > lastAddedPtr->ptr);
	/* first is <= lastAddedPtr and >= lastAddedPtr->prev */
	sortRecPtr->next = lastAddedPtr;
	sortRecPtr->prev = lastAddedPtr->prev;
	lastAddedPtr->prev->next = sortRecPtr;
	lastAddedPtr->prev = sortRecPtr;
      } else {
	/* Note lastAddedPtr will never point to the first element
	 * so prev will never be NULL.
	 */
	while (first < lastAddedPtr->ptr) {
	  lastAddedPtr = lastAddedPtr->prev;
	}
	/* first is >= lastAddedPtr and <= lastAddedPtr->next */
	sortRecPtr->next = lastAddedPtr->next;
	sortRecPtr->prev = lastAddedPtr;
	lastAddedPtr->next->prev = sortRecPtr;
	lastAddedPtr->next = sortRecPtr;
      }

      lastAddedPtr = sortRecPtr;
      sortRecPtr = &sortTable[stSize];
      stSize++;
      sortRecPtr->ptr = last;
      sortRecPtr->tiPtr = &endTiPtr[i];
      /* we known last >= first so this is a do while loop. */
      do {
	lastAddedPtr = lastAddedPtr->next;
      } while (last > lastAddedPtr->ptr);
      /* last is <= lastAddedPtr and >= lastAddedPtr->prev */
      sortRecPtr->next = lastAddedPtr;
      sortRecPtr->prev = lastAddedPtr->prev;
      lastAddedPtr->prev->next = sortRecPtr;
      lastAddedPtr->prev = sortRecPtr;
    }
  }
  sortRecPtr = &sortTable[0];
  do {
    calcTi(sortRecPtr->tiPtr, sortRecPtr->ptr, tiStatePtr);
    sortRecPtr = sortRecPtr->next;
  } while (sortRecPtr != NULL);
}

static Tcl_Obj* tiToObj(tiPtr)
     textidxSType* tiPtr;
{
    char tibuff[32];
    sprintf(tibuff, "%d.%d", tiPtr->linenum, tiPtr->offset);
    return Tcl_NewStringObj(tibuff, -1);
}

	/* ARGSUSED */
static int
Tcl_NewRegexpObjCmd(dummy, interp, objc, objv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int objc;				/* Number of arguments. */
    Tcl_Obj *CONST objv[];		/* Argument objects. */
{
    int noCase = 0;
    int indices = 0;
    int textidx = 0;
    int doAll = 0;
    int split = 0;
    int list = 0;
    int flatten = 0;
    int eval = 0;
    int prune = 0;
    int limit = -1;
    Tcl_Obj * patternObj;
    char *start, *end;
    char *string, *stringStart, *stringEnd;
    int stringLength;
    char *argStr;
    Tcl_Obj *CONST *argPtr;
    int numMatches = 0;
    int match = 0;			/* Initialization needed only to
					 * prevent compiler warning. */
    int i;
    int nameLength;
    Tcl_Obj* nameObj;
    static Tcl_Obj* emptyStrObj = NULL;
    static Tcl_Obj* negOneObj = NULL;
    static Tcl_Obj* emptyListObj = NULL;
    static Tcl_Obj* emptyTextIdxObj = NULL;
    static Tcl_Obj* emptyTextIdxListObj = NULL;
    Tcl_Obj* varObjs[NEWNSUBEXP*2];
    textidxSType startTis[NEWNSUBEXP];
    textidxSType endTis[NEWNSUBEXP];
    tiStateSType tiState;
    int maxVarObjs;
    int numValues;
    int subexpCount;
    int actualNumValues;
    int startidx = 0;
    int endidx = -1;
    Tcl_Obj** storePtr;
    Tcl_Obj* tryagain = NULL;
    int* couldHavePtr = NULL;
    int couldHave;
    int code = TCL_OK;
    static char *switches[] =
	    {"-all", "-indices", "-nocase",
	     "-list", "-flatten", "-split", "-eval", "-limit", "-prune",
	     "-textidx", "-start", "-end", "-tryagain",
	     "--", (char *) NULL};
    int switchObjc, index, length;
    Tcl_Obj *CONST *switchObjv;
    enum regexpSwitches {ALL_SWITCH, INDICES_SWITCH, NOCASE_SWITCH,
			 LIST_SWITCH, FLATTEN_SWITCH, SPLIT_SWITCH,
			 EVAL_SWITCH, LIMIT_SWITCH, PRUNE_SWITCH,
			 TEXTIDX_SWITCH, START_SWITCH, ENDIDX_SWITCH,
			 TRYAGAIN_SWITCH,
			 END_SWITCH};
    Tcl_RegExpExecResult reRes;
    int evalres;

    switchObjc = objc-1;
    switchObjv = objv+1;

    while (switchObjc > 0) {
	argStr = Tcl_GetStringFromObj(switchObjv[0], &length);
	if (*argStr != '-') {
	  break;
	}
	if (Tcl_GetIndexFromObj(interp, switchObjv[0], switches,
		"option", TCL_EXACT, &index) != TCL_OK) {
	    return TCL_ERROR;
	}
	switch (index) {
	    case ALL_SWITCH:
		doAll = 1;
		break;
	    case INDICES_SWITCH:
		textidx = 0;
		indices = 1;
		break;
	    case TEXTIDX_SWITCH:
		indices = 0;
		textidx = 1;
		break;
	    case NOCASE_SWITCH:
		noCase = 1;
		break;
	    case LIST_SWITCH:
		list = 1;
		flatten = 0;
		break;
	    case FLATTEN_SWITCH:
		flatten = 1;
		list = 0;
		break;
	    case EVAL_SWITCH:
	        eval = 1;
		break;
	    case PRUNE_SWITCH:
	        prune = 1;
		break;
	    case SPLIT_SWITCH:
		doAll = 1;
		flatten = 1;
		split = 1;
		break;
	    case LIMIT_SWITCH:
		switchObjc--;
		switchObjv++;
		if (switchObjc <= 0) {
	          Tcl_WrongNumArgs(interp, 1, objv, "-limit num ...");
		  return TCL_ERROR;
		}
		argStr = Tcl_GetStringFromObj(switchObjv[0], &length);
		if (*argStr == '-') {
	          Tcl_WrongNumArgs(interp, 1, objv, "-limit num ...");
		  return TCL_ERROR;
		}
		if (TCL_ERROR == Tcl_GetIntFromObj(interp, switchObjv[0],
						   &limit)) {
		  return TCL_ERROR;
		}
		break;
	    case START_SWITCH:
		switchObjc--;
		switchObjv++;
		if (switchObjc <= 0) {
	          Tcl_WrongNumArgs(interp, 1, objv, "-start num ...");
		  return TCL_ERROR;
		}
		argStr = Tcl_GetStringFromObj(switchObjv[0], &length);
		if (*argStr == '-') {
	          Tcl_WrongNumArgs(interp, 1, objv, "-start num ...");
		  return TCL_ERROR;
		}
		if (TCL_ERROR == Tcl_GetIntFromObj(interp, switchObjv[0],
						   &startidx)) {
		  return TCL_ERROR;
		}
		break;
	    case ENDIDX_SWITCH:
		switchObjc--;
		switchObjv++;
		if (switchObjc <= 0) {
	          Tcl_WrongNumArgs(interp, 1, objv, "-end num ...");
		  return TCL_ERROR;
		}
		argStr = Tcl_GetStringFromObj(switchObjv[0], &length);
		if (*argStr == '-') {
	          Tcl_WrongNumArgs(interp, 1, objv, "-end num ...");
		  return TCL_ERROR;
		}
		if (TCL_ERROR == Tcl_GetIntFromObj(interp, switchObjv[0],
						   &endidx)) {
		  return TCL_ERROR;
		}
		break;
	    case TRYAGAIN_SWITCH:
		switchObjc--;
		switchObjv++;
		if (switchObjc <= 0) {
	          Tcl_WrongNumArgs(interp, 1, objv, "-tryagain varname ...");
		  return TCL_ERROR;
		}
		argStr = Tcl_GetStringFromObj(switchObjv[0], &length);
		if (*argStr == '-') {
	          Tcl_WrongNumArgs(interp, 1, objv, "-tryagain varname ...");
		  return TCL_ERROR;
		}
		tryagain = switchObjv[0];
		couldHavePtr = &couldHave;
		couldHave = 0;
		break;
	    case END_SWITCH:			/* -- */
		switchObjc--;
		switchObjv++;
		goto doneWithSwitches;
	}
	switchObjc--;
	switchObjv++;
    }

    doneWithSwitches:
    if (switchObjc < 2) {
	Tcl_WrongNumArgs(interp, 1, objv,
	  "?switches? exp string ?matchVar? ?subMatchVar subMatchVar ...?");
	return TCL_ERROR;
    }
    objc = switchObjc - 2;
    /* objc is now the number of variable names to store matches in */
    argPtr = switchObjv;

    patternObj = (Tcl_Obj*)argPtr[0];
    string = Tcl_GetStringFromObj(argPtr[1], &stringLength);
    if (startidx < 0) {
      startidx = 0;
    }
    if (endidx < stringLength && endidx >= 0) {
      stringLength = endidx;
    }
    if (startidx > stringLength) {
      startidx = stringLength;
    }
    stringStart = string + startidx;
    stringEnd = string + stringLength;
    patternObj = Tcl_RegExpCompileObj(interp, patternObj, noCase);
    code = TCL_ERROR;
    maxVarObjs = 0;
    if (patternObj != NULL) {
      numMatches = 0;
      maxVarObjs = objc;
      tiState.linenum = 0;
      tiState.lptr = NULL;
      tiState.nextlptr = string;
      tiState.stringEnd = stringEnd;
      tiState.patternObj = patternObj;
      if (eval) {
	if ((objc % 2) == 1) {
	  if (list || flatten) {
	    Tcl_WrongNumArgs(interp, 1, objv,
			     "-eval ?switches? exp string matchVar matchScript");
	  } else {
	    Tcl_WrongNumArgs(interp, 1, objv,
			     "-eval ?switches? exp string ?matchVar matchScript? ?subMatchVar subMatchScript ...?");
	  }
	  return TCL_ERROR;
	}
	maxVarObjs /= 2;
      }
      if (maxVarObjs > NEWNSUBEXP+1) {
	maxVarObjs = NEWNSUBEXP+1;
      }
      if (list || flatten) {
	if (maxVarObjs != 1) {
	  if (split) {
	    Tcl_WrongNumArgs(interp, 1, objv,
			     "-split ?switches? exp string matchVar");
	  } else if (list) {
	    Tcl_WrongNumArgs(interp, 1, objv,
			     "-list ?switches? exp string matchVar");
	  } else if (flatten) {
	    Tcl_WrongNumArgs(interp, 1, objv,
			     "-flatten ?switches? exp string matchVar");
	  }
	  return TCL_ERROR;
	}
      }
      if (!emptyStrObj) {
	/* These guys are all static so they stay around forever */
	Tcl_Obj* objArray[2];
	emptyStrObj = Tcl_NewObj();
	Tcl_IncrRefCount(emptyStrObj);
	negOneObj = Tcl_NewIntObj(-1);
	objArray[0] = negOneObj;
	objArray[1] = negOneObj;
	emptyListObj = Tcl_NewListObj(2, objArray);
	Tcl_IncrRefCount(emptyListObj);
	emptyTextIdxObj = Tcl_NewStringObj("0.0", -1);
	objArray[0] = emptyTextIdxObj;
	objArray[1] = emptyTextIdxObj;
	emptyTextIdxListObj = Tcl_NewListObj(2, objArray);
	Tcl_IncrRefCount(emptyTextIdxListObj);
      }
      {
	regexp* rePtr = (regexp*)patternObj->internalRep.otherValuePtr;
	subexpCount = rePtr->subexpCount;
      }
      numValues = maxVarObjs;
      if (list || flatten) {
	numValues = subexpCount;
      }
      if (doAll) {
	char* p;
	Tcl_Obj* valPtr;

	for (i = 0; i < maxVarObjs; i++) {
	  if (eval) {
	    nameObj = (Tcl_Obj*)argPtr[(i*2)+2];
	  } else {
	    nameObj = (Tcl_Obj*)argPtr[i+2];
	  }
	  Tcl_GetStringFromObj(nameObj, &nameLength);
	  if (nameLength != 0) {
	    varObjs[i] = Tcl_NewObj();
	    Tcl_IncrRefCount(varObjs[i]);
	  } else {
	    varObjs[i] = NULL;
	  }
	}
	code = TCL_OK;
	for (p = stringStart; p <= stringEnd; ) {
	  if (numMatches == limit) {
	    break;
	  }
	  match = Tcl_RegExpExecObj(interp, patternObj,
				    p, stringStart, stringEnd - p, &reRes,
				    couldHavePtr);
	  if (match < 0) {
	    code = TCL_ERROR;
	    break;
	  }
	  if (!match) {
	    if (couldHavePtr && couldHavePtr[0]) {
		if (!Tcl_ObjSetVar2(interp, tryagain, NULL,
				    Tcl_NewIntObj(p - string),
				    TCL_PARSE_PART1|TCL_LEAVE_ERR_MSG)) {
		  code = TCL_ERROR;
		  goto ERROR_RETURN;
		}
		/* Bump p to make sure split doesn't eat the remainder */
		p = stringEnd + 1; 
	    }
	    break;
	  }
	  numMatches++;
	  if (eval) {
	    for (i = 0; i < maxVarObjs; i++) {
	      if (varObjs[i] != NULL) {
		Tcl_DecrRefCount(varObjs[i]);
		varObjs[i] = Tcl_NewObj();
		Tcl_IncrRefCount(varObjs[i]);
	      }
	    }
	  }
	  if (split && varObjs[0] != NULL) {
	    start = REGEXP_RANGE0_START(reRes);
	    if (indices) {
	      Tcl_ListObjAppendElement(interp, varObjs[0],
				       Tcl_NewIntObj((int)(p - string)));
	      Tcl_ListObjAppendElement(interp, varObjs[0],
				       Tcl_NewIntObj((int)(start - string - 1)));
	    } else if (textidx) {
	      textidxSType ti;
	      calcTi(&ti, p, &tiState);
	      Tcl_ListObjAppendElement(interp, varObjs[0], tiToObj(&ti));
	      calcTi(&ti, start, &tiState);
	      Tcl_ListObjAppendElement(interp, varObjs[0], tiToObj(&ti));
	    } else {
	      int length = (start - p);
	      if (length == 0) {
		valPtr = emptyStrObj;
	      } else {
		valPtr = Tcl_NewStringObj(p, length);
	      }
	      Tcl_ListObjAppendElement(interp, varObjs[0], valPtr);
	    }
	  }
	  storePtr = varObjs;
	  if (list && numValues > split) {
	    storePtr++;
	  }
	  if (textidx) {
	    calcAllVarTis(startTis, endTis, &tiState, &reRes, subexpCount);
	  }
	  actualNumValues = numValues;
	  for (i = split; i < numValues; i++) {
	    if (*storePtr != NULL || (list && varObjs[0] != NULL)) {
	      Tcl_RegExpRangeObj(patternObj, &reRes, i, &start, &end);
	      if (prune && start == NULL) {
		actualNumValues--;
	      } else if (indices) {
		if (start == NULL) {
		  if (list) {
		    *storePtr++ = negOneObj;
		    *storePtr++ = negOneObj;
		  } else {
		    Tcl_ListObjAppendElement(interp, *storePtr, negOneObj);
		    Tcl_ListObjAppendElement(interp, *storePtr, negOneObj);
		  }
		} else {
		  if (list) {
		    *storePtr++ = Tcl_NewIntObj((int)(start - string));
		    *storePtr++ = Tcl_NewIntObj((int)(end - string - 1));
		  } else {
		    Tcl_ListObjAppendElement(interp, *storePtr,
					     Tcl_NewIntObj((int)(start - string)));
		    Tcl_ListObjAppendElement(interp, *storePtr,
					     Tcl_NewIntObj((int)(end - string - 1)));
		  }
		}
	      } else if (textidx) {
		if (start == NULL) {
		  if (list) {
		    *storePtr++ = emptyTextIdxObj;
		    *storePtr++ = emptyTextIdxObj;
		  } else {
		    Tcl_ListObjAppendElement(interp, *storePtr, emptyTextIdxObj);
		    Tcl_ListObjAppendElement(interp, *storePtr, emptyTextIdxObj);
		  }
		} else {
		  if (list) {
		    *storePtr++ = tiToObj(&startTis[i]);
		    *storePtr++ = tiToObj(&endTis[i]);
		  } else {
		    Tcl_ListObjAppendElement(interp, *storePtr,
					     tiToObj(&startTis[i]));
		    Tcl_ListObjAppendElement(interp, *storePtr,
					     tiToObj(&endTis[i]));
		  }
		}
	      } else {
		if (start == NULL) {
		  valPtr = emptyStrObj;
		} else {
		  int length = (end - start);
		  if (length == 0) {
		    valPtr = emptyStrObj;
		  } else {
		    valPtr = Tcl_NewStringObj(start, length);
		  }
		}
		if (list) {
		  *storePtr++ = valPtr;
		} else {
		  Tcl_ListObjAppendElement(interp, *storePtr, valPtr);
		}
	      }
	    }
	    if (!list && !flatten) {
	      storePtr++;
	    }
	  }
	  if (list && varObjs[0] != NULL) {
	    int lsize = actualNumValues;
	    if (split) {
	      lsize--;
	    }
	    if (indices || textidx) {
	      lsize *= 2;
	    }
	    if (lsize > 0) {
	      Tcl_ListObjAppendElement(interp, varObjs[0],
				       Tcl_NewListObj(lsize, &varObjs[1]));
	    }
	  }
	  if (eval) {
	    /* first set any vars */
	    for (i = 0; i < maxVarObjs; i++) {
	      if (varObjs[i] != NULL) {
		nameObj = (Tcl_Obj*)argPtr[(i*2)+2];
		if (!Tcl_ObjSetVar2(interp, nameObj, NULL, varObjs[i],
				    TCL_PARSE_PART1|TCL_LEAVE_ERR_MSG)) {
		  code = TCL_ERROR;
		  goto ERROR_RETURN;
		}
	      }
	    }
	    /* now evaluate each of the scripts */
	    for (i = 0; i < maxVarObjs; i++) {
	      Tcl_RegExpRangeObj(patternObj, &reRes, i, &start, &end);
	      if (start != NULL) {
		/* only evaluate if a match was found */
		Tcl_Obj* scriptObj = (Tcl_Obj*)argPtr[(i*2)+3];
		evalres = Tcl_EvalObj(interp, scriptObj);
		switch (evalres) {
		case TCL_OK:
		  break;
		case TCL_BREAK:
		  /* TCL_BREAK means to act as if no match occurred and quit. */
		  numMatches -= 1;
		  code = TCL_OK;
		  goto OK_RETURN;
		case TCL_CONTINUE:
		  /* TCL_CONTINUE means to act as if no match occurred
		   * and try again at the next position.
		   */
		  numMatches -= 1;
		  p = start;
		  end = start;
		  goto DO_CONTINUE;
		case TCL_ERROR:
		  /* TCL_ERROR means that something went wrong. */
		  /* Need to set a special error message? */
		case TCL_RETURN:
		default:
		  /* TCL_RETURN or anything else means to clean up and pass the
		   * code back to the caller.
		   * Note that in this case the result variable is not set.
		   */
		  code = evalres;
		  goto ERROR_RETURN;
		}
	      }
	    }
	  }
	  end = REGEXP_RANGE0_END(reRes);
	DO_CONTINUE:
	  if (end == p) {
	    /*
	     * Always consume at least one character of the input string
	     * in order to prevent infinite loops.
	     */
	    p = end + 1;
	  } else {
	    p = end;
	  }
	} /* end for */
	if (split && code == TCL_OK) {
	  if (p <= stringEnd && varObjs[0] != NULL) {
	    int length = (stringEnd - p);
	    if (numMatches > 0 || length > 0) {
	      if (eval) {
		varObjs[0] = Tcl_NewObj();
		Tcl_IncrRefCount(varObjs[0]);
	      }
	      if (indices) {
		Tcl_ListObjAppendElement(interp, varObjs[0],
					 Tcl_NewIntObj((int)(p - string)));
		Tcl_ListObjAppendElement(interp, varObjs[0],
					 Tcl_NewIntObj((int)(stringEnd - string - 1)));
	      } else if (textidx) {
		textidxSType ti;
		calcTi(&ti, p, &tiState);
		Tcl_ListObjAppendElement(interp, varObjs[0], tiToObj(&ti));
		calcTi(&ti, stringEnd, &tiState);
		Tcl_ListObjAppendElement(interp, varObjs[0], tiToObj(&ti));
	      } else {
		if (length == 0) {
		  valPtr = emptyStrObj;
		} else {
		  valPtr = Tcl_NewStringObj(p, length);
		}
		Tcl_ListObjAppendElement(interp, varObjs[0], valPtr);
	      }
	      if (eval) {
		nameObj = (Tcl_Obj*)argPtr[2];
		if (!Tcl_ObjSetVar2(interp, nameObj, NULL, varObjs[0],
				    TCL_PARSE_PART1|TCL_LEAVE_ERR_MSG)) {
		  code = TCL_ERROR;
		  goto ERROR_RETURN;
		}
		evalres = Tcl_EvalObj(interp, (Tcl_Obj*)argPtr[3]);
		switch (evalres) {
		case TCL_BREAK:
		case TCL_CONTINUE:
		case TCL_OK:
		  break;
		case TCL_ERROR:
		  /* TCL_ERROR means that something went wrong. */
		  /* Need to set a special error message? */
		case TCL_RETURN:
		default:
		  /* TCL_RETURN or anything else means to clean up and pass the
		   * code back to the caller.
		   * Note that in this case the result variable is not set.
		   */
		  code = evalres;
		  goto ERROR_RETURN;
		}
	      }
	    }
	  }
	}
      } else { /* not doAll */
	for (i = 0; i < maxVarObjs; i++) {
	  varObjs[i] = NULL;
	}
	match = Tcl_RegExpExecObj(interp, patternObj,
				  stringStart, stringStart,
				  stringEnd - stringStart, &reRes,
				  couldHavePtr);
	if (match >= 0) {
	  code = TCL_OK;
	  if (match == 0) {
	    if (couldHavePtr && couldHavePtr[0]) {
	      char *tryAgainPtr = stringStart;
	      if (!split && couldHavePtr[0] != 1) {
		/* the regexp engine already skipped over some of the
		 * starting bytes.
		 */
		tryAgainPtr = (char*)couldHavePtr[0];
	      }
	      if (!Tcl_ObjSetVar2(interp, tryagain, NULL,
				  Tcl_NewIntObj(tryAgainPtr - string),
				  TCL_PARSE_PART1|TCL_LEAVE_ERR_MSG)) {
		code = TCL_ERROR;
		goto ERROR_RETURN;
	      }
	    }
	  } else {
	    numMatches++;
	    storePtr = varObjs;
	    if (list || flatten) {
	      storePtr++;
	    }
	    if (textidx) {
	      calcAllVarTis(startTis, endTis, &tiState, &reRes, subexpCount);
	    }
	    for (i = 0; i < numValues; i++) {
	      if (eval) {
		nameObj = (Tcl_Obj*)argPtr[(i*2)+2];
	      } else {
		nameObj = (Tcl_Obj*)argPtr[i+2];
	      }
	      Tcl_GetStringFromObj(nameObj, &nameLength);
	      if (nameLength != 0 || list || flatten) {
		Tcl_RegExpRangeObj(patternObj, &reRes, i, &start, &end);
		if (start == NULL) {
		  if (indices) {
		    *storePtr = emptyListObj;
		  } else if (textidx) {
		    *storePtr = emptyTextIdxListObj;
		  } else {
		    *storePtr = emptyStrObj;
		  }
		} else {
		  if (indices) {
		    Tcl_Obj* objArray[2];
		    objArray[0] = Tcl_NewIntObj((int)(start - string));
		    objArray[1] = Tcl_NewIntObj((int)(end - string - 1));
		    *storePtr = Tcl_NewListObj(2, objArray);
		  } else if (textidx) {
		    Tcl_Obj* objArray[2];
		    objArray[0] = tiToObj(&startTis[i]);
		    objArray[1] = tiToObj(&endTis[i]);
		    *storePtr = Tcl_NewListObj(2, objArray);
		  } else {
		    int length = (end - start);
		    if (length == 0) {
		      *storePtr = emptyStrObj;
		    } else {
		      *storePtr = Tcl_NewStringObj(start, length);
		    }
		  }
		}
		if (!list && !flatten) {
		  Tcl_IncrRefCount(*storePtr);
		}
	      }
	      storePtr++;
	    }
	    if (list || flatten) {
	      varObjs[0] = Tcl_NewListObj(numValues, &varObjs[1]);
	      Tcl_IncrRefCount(varObjs[0]);
	    }
	    if (eval) {
	      /* first set any vars */
	      for (i = 0; i < maxVarObjs; i++) {
		if (varObjs[i] != NULL) {
		  nameObj = (Tcl_Obj*)argPtr[(i*2)+2];
		  if (!Tcl_ObjSetVar2(interp, nameObj, NULL, varObjs[i],
				      TCL_PARSE_PART1|TCL_LEAVE_ERR_MSG)) {
		    code = TCL_ERROR;
		    goto ERROR_RETURN;
		  }
		}
	      }
	      /* now evaluate each of the scripts */
	      for (i = 0; i < maxVarObjs; i++) {
		Tcl_RegExpRangeObj(patternObj, &reRes, i, &start, &end);
		if (start != NULL) {
		  /* only evaluate if a match was found */
		  Tcl_Obj* scriptObj = (Tcl_Obj*)argPtr[(i*2)+3];
		  evalres = Tcl_EvalObj(interp, scriptObj);
		  switch (evalres) {
		  case TCL_CONTINUE:
		  case TCL_OK:
		    break;
		  case TCL_BREAK:
		    /* TCL_BREAK means do no more evals. */
		    code = TCL_OK;
		    goto OK_RETURN;
		  case TCL_ERROR:
		    /* TCL_ERROR means that something went wrong. */
		    /* Need to set a special error message? */
		  case TCL_RETURN:
		  default:
		    /* TCL_RETURN or anything else means to clean up and pass the
		     * code back to the caller.
		     * Note that in this case the result variable is not set.
		     */
		    code = evalres;
		    goto ERROR_RETURN;
		  }
		}
	      }
	    }
	  }
	}
      }
    }
    if (code == TCL_OK && !eval && (numMatches > 0 || split)) {
      Tcl_Obj* emptyObj;
      /*
       * If additional variable names have been specified, return
       * index information in those variables.
       */
      if (indices) {
	if (doAll) {
	  emptyObj = negOneObj;
	} else {
	  emptyObj = emptyListObj;
	}
      } else if (textidx) {
	if (doAll) {
	  emptyObj = emptyTextIdxObj;
	} else {
	  emptyObj = emptyTextIdxListObj;
	}
      } else {
	emptyObj = emptyStrObj;
      }

      for (i = 0; i < objc; i++) {
	Tcl_Obj* valueObj;
	int j = i;
	nameObj = (Tcl_Obj*)argPtr[i+2];
	Tcl_GetStringFromObj(nameObj, &nameLength);
	if (nameLength != 0) {
	  if (j >= maxVarObjs) {
	    j = maxVarObjs-1;
	  }
	  valueObj = varObjs[j];
	  if (valueObj == NULL) {
	    valueObj = emptyObj;
	  }
	  if (!Tcl_ObjSetVar2(interp, nameObj, NULL, valueObj,
			      TCL_PARSE_PART1|TCL_LEAVE_ERR_MSG)) {
	    code = TCL_ERROR;
	    break;
	  }
	}
      }
    }

 OK_RETURN:
    if (code == TCL_OK) {
      if (couldHavePtr && !couldHavePtr[0]) {
	if (!Tcl_ObjSetVar2(interp, tryagain, NULL, negOneObj,
			    TCL_PARSE_PART1|TCL_LEAVE_ERR_MSG)) {
	  code = TCL_ERROR;
	  goto ERROR_RETURN;
	}
      }
      Tcl_SetObjResult(interp, Tcl_NewIntObj(numMatches));
    }

 ERROR_RETURN:
    /* Free objects */
    for (i = 0; i < maxVarObjs; i++) {
      if (varObjs[i] != NULL) {
	Tcl_DecrRefCount(varObjs[i]);
      }
    }
    return code;
}

static int
Tcl_NewRegDumpObjCmd(dummy, interp, objc, objv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int objc;				/* Number of arguments. */
    Tcl_Obj *CONST objv[];		/* Argument objects. */
{
    char buff[80];
    int alreadyRegexp;
    Tcl_Obj * patternObj;
    Tcl_Obj *resObjPtr;

    if (objc != 2) {
	Tcl_WrongNumArgs(interp, 1, objv, "exp");
	return TCL_ERROR;
    }
    patternObj = (Tcl_Obj*)objv[1];
    alreadyRegexp = (patternObj->typePtr == &tclRegExpType);
    patternObj = Tcl_RegExpCompileObj(interp, patternObj, 0);
    if (patternObj == NULL) {
      return TCL_ERROR;
    }
    if (TclNewRegDump(interp, (regexp*)patternObj->internalRep.otherValuePtr)
	== TCL_ERROR) {
      return TCL_ERROR;
    }
    resObjPtr = Tcl_GetObjResult(interp);
    sprintf(buff, "exp's object %p%s refcount: %d",
	    patternObj,
	    alreadyRegexp ? "" : ((lastFoundInCache == patternObj) ?
				  " found in cache" : " needed compilation"),
	    patternObj->refCount);

    return Tcl_ListObjAppendElement(interp, resObjPtr,
				    Tcl_NewStringObj(buff, -1));
}

/* MOVE_FROM_TO_STOPAT copies bytes starting at 'from' and ending at they byte
 * before 'end'. They are copied to 'toPtr'.
 * toPtr is modified to point to the byte after the last.
 */
#define MOVE_FROM_TO_STOPAT(from, toPtr, end) \
	      { \
		char* mmsrc = from; \
	        while (mmsrc < end) { \
		  *toPtr++ = *mmsrc++;\
	        }\
	      }


/* Backrefs will be trashed in subObj is copied to any byte at or after start.
 * However if start, gapStart, and subObj all point to the same place backrefs
 * are safe.
 */
#define backrefsSafe(subObj, subLen, gapStart, start) \
  (((gapStart + subLen) <= start) || \
   ((gapStart == start) && (gapStart == subObj)))

static int okToTrash(ssPtr, ssEnd, patternObj, reResPtr, endTrashed)
long *ssPtr;
long *ssEnd;
Tcl_Obj* patternObj;
Tcl_RegExpExecResult *reResPtr;
char *endTrashed;
{
  while (ssPtr < ssEnd) {
    long v = *ssPtr++;
    if (v < NEWNSUBEXP) {
      if (v >= 0) {
	char* subStart;
	char* subEnd;
	Tcl_RegExpRangeObj(patternObj, reResPtr, v, &subStart, &subEnd);
	if ((subStart != NULL) && (subEnd != NULL)) {
	  if ((subEnd - subStart) > 0) {
	    if (subStart < endTrashed) {
	      /* This area of matched is still needed. */
	      return 0;
	    }
	  }
	}
      }
    } else {
      ssPtr++;
    }
  }
  return 1;
}

static int
parseSubSpec(string, strSize, stackBuffPtr, buffSize)
    char* string;
    long strSize;
    long** stackBuffPtr;
    int  buffSize;
{
  int subSpecSize = 0;
  long* subSpecBuffPtr = *stackBuffPtr;
  char *stringEnd = string + strSize;
  char *firstChar;
  char *p;
  char c;
  int inRun;

  inRun = 0;
  for (p = string; p != stringEnd; p++) {
    c = *p;
    if (c == '&') {
      inRun = 0;
      subSpecSize++;
    } else if (c == '\\') {
      c = p[1];
      if (((c >= '0') && (c <= '9')) || (c == '\\') || (c == '&')) {
	/* a subExpression or a literal \ or & */
	inRun = 0;
	subSpecSize++;
	p++;
      } else if (!inRun) {
	/* a run of literal charactes */
	inRun = 1;
	subSpecSize += 2;
      }
    } else if (!inRun) {
      /* a run of literal charactes */
      inRun = 1;
      subSpecSize += 2;
    }
  }
  if (subSpecSize > buffSize) {
    subSpecBuffPtr = (long*)ckalloc(subSpecSize * sizeof(long));
    *stackBuffPtr = subSpecBuffPtr;
  }
    
  if (subSpecSize > 0) {
    for (p = firstChar = string; p != stringEnd; p++) {
      int index;

      c = *p;
      if (c == '&') {
	index = 0;
      } else if (c == '\\') {
	c = p[1];
	if ((c >= '0') && (c <= '9')) {
	  index = c - '0';
	} else if (c == '\\') {
	  index = -1;
	} else if (c == '&') {
	  index = -2;
	} else {
	  continue;
	}
      } else {
	continue;
      }
      if (firstChar != p) {
	/* store the length */
	*subSpecBuffPtr++ = (p - firstChar) + NEWNSUBEXP;
	/* store offset to first character */
	*subSpecBuffPtr++ = firstChar - string;
      }
      *subSpecBuffPtr++ = index;
      if (*p == '\\') {
	p++;
      }
      firstChar = p+1;
    }
    if (firstChar != p) {
      /* store the length */
      *subSpecBuffPtr++ = (stringEnd - firstChar) + NEWNSUBEXP;
      /* store offset to first character */
      *subSpecBuffPtr++ = firstChar - string;
    }
  }
  return subSpecSize;
}
/*
 *----------------------------------------------------------------------
 *
 * Tcl_NewRegsubCmd --
 *
 *	This procedure is invoked to process the "nresub" Tcl command.
 *	See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */
	/* ARGSUSED */
static int
Tcl_NewRegsubObjCmd(dummy, interp, objc, objv)
    ClientData dummy;			/* Not used. */
    Tcl_Interp *interp;			/* Current interpreter. */
    int objc;				/* Number of arguments. */
    Tcl_Obj *CONST objv[];		/* Argument objects. */
{
    int noCase = 0, all = 0;
    Tcl_Obj * patternObj;
    char *p;
    int match, numMatches;
    char *start, *end, *subStart, *subEnd;
    Tcl_Obj *subObj = NULL;
    char *string, *stringEnd;
    int stringLength;
    Tcl_Obj *CONST *argPtr;
    static char *switches[] =
	    {"-all", "-nocase", "-eval", "--", (char *) NULL};
    int switchObjc, index, length;
    Tcl_Obj *CONST *switchObjv;
    enum regsubSwitches {ALL_SWITCH, NOCASE_SWITCH, EVAL_SWITCH, END_SWITCH};
    Tcl_RegExpExecResult reRes;
    int eval = 0;
    Tcl_Obj *evalValueObj = NULL;
    Tcl_Obj *matchStringArray[NEWNSUBEXP];
    int numMatchStrings;
    int evalSetFlags;
    int i;
    static Tcl_Obj *emptyStrObj = NULL;
    int doContinue;
    Tcl_Obj *evalNameObj = NULL;
#define SSBUF_SIZE 16
    long subSpecBuff[SSBUF_SIZE];
    int subSpecLength = 0;
    long* subSpecPtr = NULL;
    long* subSpecPtrEnd = NULL;
    char* subSpecStr = NULL;
    int freeSubSpecBuff = 0;
    int inplace = 0;
    int justCount = 0; /* true if regsub being used to just count num matches */
    int alreadySet = 0;
    char* gapStart = NULL;
    char* gapEnd = NULL;

    switchObjc = objc-1;
    switchObjv = objv+1;

    while (switchObjc > 0) {
	char *argStr = Tcl_GetStringFromObj(switchObjv[0], &length);
	if (*argStr != '-') {
	  break;
	}
	if (Tcl_GetIndexFromObj(interp, switchObjv[0], switches,
		"option", TCL_EXACT, &index) != TCL_OK) {
	    return TCL_ERROR;
	}
	switch (index) {
	    case ALL_SWITCH:
		all = 1;
		break;
	    case NOCASE_SWITCH:
		noCase = 1;
		break;
	    case EVAL_SWITCH:
		eval = 1;
		break;
	    case END_SWITCH:			/* -- */
		switchObjc--;
		switchObjv++;
		goto doneWithSwitches;
	}
	switchObjc--;
	switchObjv++;
    }

    doneWithSwitches:
    if (eval) {
      if (switchObjc != 5) {
	Tcl_WrongNumArgs(interp, 1, objv,
			 "-eval ?switches? exp string subSpec varName evalName");
	return TCL_ERROR;
      }
    } else {
      if (switchObjc != 4) {
	Tcl_WrongNumArgs(interp, 1, objv,
			 "?switches? exp string subSpec varName");
	return TCL_ERROR;
      }
    }
    objc = switchObjc;
    argPtr = switchObjv;

    patternObj = (Tcl_Obj*)argPtr[0];
    string = Tcl_GetStringFromObj(argPtr[1], &stringLength);
    stringEnd = string + stringLength;

    patternObj = Tcl_RegExpCompileObj(interp, patternObj, noCase);
    if (patternObj == NULL) {
      return TCL_ERROR;
    }

    {
      /* see if we are just counting the number of matches */
      int resultVarNameLen = 0;
      (void)Tcl_GetStringFromObj(argPtr[3], &resultVarNameLen);
      justCount = (resultVarNameLen == 0);
    }
    if (eval) {
      /* Determine the size of the match list. We know the size of the list
       * will always be the number of subexpression in the pattern.
       */
      char *evalName;
      int evalNameLen;
      regexp* rePtr = (regexp*)patternObj->internalRep.otherValuePtr;
      evalNameObj = argPtr[4];
      numMatchStrings = rePtr->subexpCount;
      evalSetFlags = TCL_LEAVE_ERR_MSG;
      /* Figure out if we need to do a PARSE_PART1 every time we set the var */
      evalName = Tcl_GetStringFromObj(evalNameObj, &evalNameLen);
      if (evalNameLen == 0) {
	evalNameObj = NULL;
      } else if (memchr(evalName, '(', evalNameLen) != NULL) {
	/* The evalName contains a '(' so its probably an array name */
	evalSetFlags |= TCL_PARSE_PART1;
      }
    } else if (!justCount) {
      int subSpecStrLen;
      subSpecStr = Tcl_GetStringFromObj(argPtr[2], &subSpecStrLen);
      subSpecPtr = subSpecBuff;
      subSpecLength = parseSubSpec(subSpecStr, subSpecStrLen,
				   &subSpecPtr, SSBUF_SIZE);
      subSpecPtrEnd = &subSpecPtr[subSpecLength];
      if (subSpecPtr != subSpecBuff) {
	freeSubSpecBuff = 1;
      }
    }
    if (justCount) {
      alreadySet = 1;
    } else {
      /* In-place substitions can be done if:
	 -- refcount of argPtr[1] is 1 (or 2 and one is argPtr[3] from argPtr3
	 If count is 2 then and inplace used then no need for assignment at end
	 -- -eval is used or subSpec contains no backreferences including &.
	 Backrefs can be ok if they completely match the text to be
	 substituted. For example {&} is always ok, {\1} is ok if \1 == &.
	 They are also ok if they match nothing.
	 We are lazy about figuring out if the backrefs are ok.
	 -- the length of the new text is <= the length of the replaced text
	 Can be lazy about figuring this out
      */
      if (argPtr[1]->refCount == 1) {
	inplace = 1;
      } else if (argPtr[1]->refCount == 2) {
	Tcl_Obj* dstObjPtr;
	dstObjPtr = Tcl_ObjGetVar2(interp, argPtr[3], NULL, TCL_PARSE_PART1);
	if (dstObjPtr == argPtr[1]) {
	  inplace = 1;
	  alreadySet = 1;
	}
      }
      if (inplace) {
	subObj = NULL;
	gapStart = string;
	gapEnd = gapStart;
      } else {
	subObj = Tcl_NewObj();
	if (stringLength > 0) {
	  /* pregrow the target object so we don't need to keep growing
	   * and copying it.
	   */
	  Tcl_SetObjLength(subObj, stringLength + (stringLength / 2));
	  Tcl_SetObjLength(subObj, 0);
	}
      }
    }

    /*
     * The following loop is to handle multiple matches within the
     * same source string;  each iteration handles one match and its
     * corresponding substitution.  If "-all" hasn't been specified
     * then the loop body only gets executed once.
     */

    numMatches = 0;
    for (p = string; p <= stringEnd; ) {
      doContinue = 0;
      match = Tcl_RegExpExecObj(interp, patternObj, p, string, stringEnd - p,
				&reRes, NULL);
      if (match < 0) {
	if (subObj) {
	  Tcl_DecrRefCount(subObj);
	}
	if (freeSubSpecBuff) {
	  ckfree((char*)subSpecPtr);
	}
	return TCL_ERROR;
      }
      if (!match) {
	break;
      }
      numMatches += 1;

      /*
       * Copy the portion of the source string before the match to the
       * result variable.
       */

      start = REGEXP_RANGE0_START(reRes);
      end = REGEXP_RANGE0_END(reRes);
      if (eval) {
	int evalres;
	int alldone = 0;
	if (subObj) {
	  Tcl_AppendToObj(subObj, p, start - p);
	}
	if (evalNameObj) {
	  /* Create strings for all the matched expressions */
	  for (i = 0; i < numMatchStrings; i++) {
	    int len;
	    Tcl_RegExpRangeObj(patternObj, &reRes, i, &subStart, &subEnd);
	    len = subEnd - subStart;
	    if (len > 0) {
	      matchStringArray[i] = Tcl_NewStringObj(subStart, len);
	    } else {
	      if (emptyStrObj == NULL) {
		emptyStrObj = Tcl_NewObj();
		Tcl_IncrRefCount(emptyStrObj);
	      }
	      matchStringArray[i] = emptyStrObj;
	    }
	  }
	  if (numMatchStrings == 1) {
	    /* no subexpressions so don't bother with a list */
	    evalValueObj = matchStringArray[0];
	  } else {
	    evalValueObj = Tcl_NewListObj(numMatchStrings, matchStringArray);
	  }
	  if (!Tcl_ObjSetVar2(interp, evalNameObj, NULL, evalValueObj,
			      evalSetFlags)) {
	    if (subObj) Tcl_DecrRefCount(subObj);
	    Tcl_DecrRefCount(evalValueObj);
	    return TCL_ERROR;
	  }
	}
	/* Now evaluate the subSpec code which can have references to
	 * the evalName variable.
	 */
	evalres = Tcl_EvalObj(interp, argPtr[2]);
	switch (evalres) {
	case TCL_RETURN:
	case TCL_OK: if (!justCount) {
	  int resLength;
	  Tcl_Obj* resObj = Tcl_GetObjResult(interp);
	  char* resStr = Tcl_GetStringFromObj(resObj, &resLength);
	  /* TCL_OK and TCL_RETURN means to take the result string
	   * and append it.
	   */
	  if (!subObj) {
	    int matchLength = end - start;
	    if (matchLength > 0) {
	      if (gapStart != gapEnd) {
		/* A previous deleted range exists */
		/* need to move the bytes between the two ranges down */
#ifdef HAS_MEMMOVE
		int moveLength = start - gapEnd;
		if (moveLength > 7) {
		  memmove(gapStart, gapEnd, moveLength);
		  gapStart += moveLength;
		} else {
		  MOVE_FROM_TO_STOPAT(gapEnd, gapStart, start);
		}
#else
		MOVE_FROM_TO_STOPAT(gapEnd, gapStart, start);
#endif
	      } else {
		gapStart = start;
	      }
	      gapEnd = end;
	      /* At this point we only have one gap and it includes the
	       * currently matched text.
	       * 'start' and 'end' still mark the current match.
	       */
	    }
	    if (resLength <= (gapEnd - gapStart)) {
	      if (resLength > 0) {
		memcpy(gapStart, resStr, resLength);
		gapStart += resLength;
	      }
	      break;
	    } else {
	      /* couldn't do the inplace because we ran out of room */
	      alreadySet = 0;
	      subObj = Tcl_NewObj();
	      if (stringLength > 0) {
		/* pregrow the target object so we don't need to keep growing
		 * and copying it.
		 */
		Tcl_SetObjLength(subObj, stringLength + (stringLength / 2));
		Tcl_SetObjLength(subObj, 0);
	      }
	      if (gapStart == gapEnd) {
		gapStart = start;
	      }
	      Tcl_AppendToObj(subObj, string, gapStart - string);
	      gapStart = NULL;
	      gapEnd = NULL;
	    }
	  }
	  Tcl_AppendToObj(subObj, resStr, resLength);
	  break;
	}
	case TCL_ERROR:
	  /* TCL_ERROR means that something went wrong. */
	  /* Need to set a special error message? */
	  if (subObj) Tcl_DecrRefCount(subObj);
	  return TCL_ERROR;
	case TCL_BREAK:
	  /* TCL_BREAK means to act as if no match occurred and quit. */
	  numMatches -= 1;
	  p = start;
	  alldone = 1;
	  break;
	case TCL_CONTINUE:
	  /* TCL_CONTINUE means to act as if no match occurred
	   * and try again at the next position.
	   */
	  numMatches -= 1;
	  p = start;
	  end = start;
	  doContinue = 1;
	  break;
	default:
	  /* anything else means to clean up and pass the
	   * code back to the caller.
	   * Note that in this case the result variable is not set.
	   */
	  if (subObj) Tcl_DecrRefCount(subObj);
	  return evalres;
	}
	if (alldone) {
	  break;
	}
      } else if (!justCount) {
	/* Not an eval */
	int backrefok = 1;
	long* ssPtr = subSpecPtr;
	if (subObj) {
	  Tcl_AppendToObj(subObj, p, start - p);
	} else {
	  int matchLength = end - start;
	  if (matchLength > 0) {
	    if (gapStart != gapEnd) {
	      /* A previous deleted range exists */
	      /* need to move the bytes between the two ranges down */
#ifdef HAS_MEMMOVE
	      int moveLength = start - gapEnd;
	      if (moveLength > 7) {
		memmove(gapStart, gapEnd, moveLength);
		gapStart += moveLength;
	      } else {
		MOVE_FROM_TO_STOPAT(gapEnd, gapStart, start);
	      }
#else
	      MOVE_FROM_TO_STOPAT(gapEnd, gapStart, start);
#endif
	    } else {
	      gapStart = start;
	    }
	    gapEnd = end;
	    /* At this point we only have one gap and it includes the
	     * currently matched text.
	     * 'start' and 'end' still mark the current match.
	     */
	  }
	}
	/* Iterate over the compiled subspec appending to subObj */
	while (ssPtr < subSpecPtrEnd) {
	  long v = *ssPtr++;
	  int subLen;
	  if (v < NEWNSUBEXP) {
	    if (v >= 0) {
	      Tcl_RegExpRangeObj(patternObj, &reRes, v, &subStart, &subEnd);
	      subLen = 0;
	      if ((subStart != NULL) && (subEnd != NULL)) {
		subLen = subEnd - subStart;
	      }
	    } else if (v == -1) {
	      subStart = "\\";
	      subLen = 1;
	    } else if (v == -2) {
	      subStart = "&";
	      subLen = 1;
	    } else {
	      Tcl_SetResult(interp, "internal error in subspec code",
			    TCL_STATIC);
	      if (freeSubSpecBuff) {
		ckfree((char*)subSpecPtr);
	      }
	      return TCL_ERROR;
	    }
	  } else {
	    subLen = v - NEWNSUBEXP;
	    subStart = subSpecStr;
	    subStart += *ssPtr++;
	  }
	  if (subLen > 0) {
	    if (subObj) {
	      Tcl_AppendToObj(subObj, subStart, subLen);
	    } else {
	      int gapLen = gapEnd - gapStart;
	      if ((subLen <= gapLen) &&
		  (backrefsSafe(subStart, subLen, gapStart, start) ||
		   okToTrash(ssPtr, subSpecPtrEnd, patternObj, &reRes,
			     gapStart+subLen))) {
		memcpy(gapStart, subStart, subLen);
		gapStart += subLen;
	      } else {
		/* couldn't do the inplace because we ran out of room */
		alreadySet = 0;
		subObj = Tcl_NewObj();
		if (stringLength > 0) {
		  /* pregrow the target object so we don't need to keep growing
		   * and copying it.
		   */
		  Tcl_SetObjLength(subObj, stringLength + (stringLength / 2));
		  Tcl_SetObjLength(subObj, 0);
		}
		Tcl_AppendToObj(subObj, string, gapStart - string);
		Tcl_AppendToObj(subObj, subStart, subLen);
		gapStart = NULL;
		gapEnd = NULL;
	      }
	    }
	  }
	}
      }
      if (end == p) {
	/*
	 * Always consume at least one character of the input string
	 * in order to prevent infinite loops.
	 */

	if (p < stringEnd && subObj) {
	  Tcl_AppendToObj(subObj, p, 1);
	}
	p++;
      } else {
	p = end;
      }
      if (!all && !doContinue) {
	break;
      }
    }

    /*
     * Copy the portion of the source string after the last match to the
     * result variable.
     */
    if ((p < stringEnd) && subObj) {
      Tcl_AppendToObj(subObj, p, stringEnd - p);
    }
    if (!subObj && gapStart != gapEnd) {
#ifdef HAS_MEMMOVE
      int moveLength = stringEnd - gapEnd;
      if (moveLength > 0) {
	memmove(gapStart, gapEnd, moveLength);
	gapStart += moveLength;
      }
#else
      MOVE_FROM_TO_STOPAT(gapEnd, gapStart, stringEnd);
#endif
      if (gapStart != stringEnd) {
	/* Note that Tcl_SetObjLength is not used because it does not
	 * work with objects whose refcount > 1.
	 */
	stringEnd = gapStart;
	*stringEnd = '\0';
	argPtr[1]->length = stringEnd - string;
      }
    }
    if (freeSubSpecBuff) {
      ckfree((char*)subSpecPtr);
    }

    if (!subObj && inplace) {
      Tcl_Obj* ipPtr = argPtr[1];
      Tcl_ObjType *ipTypePtr = ipPtr->typePtr;
      /* The object string rep was changed in place so need to
       * invalidate any internal representation.
       */
      if ((ipTypePtr != NULL) && (ipTypePtr->freeIntRepProc != NULL)) {
	ipTypePtr->freeIntRepProc(ipPtr);
      }
      ipPtr->typePtr = NULL;
    }
    if (!alreadySet) {
      Tcl_Obj* setObj = subObj;
      if (!setObj) {
	setObj = argPtr[1];
      }
      if (!Tcl_ObjSetVar2(interp, argPtr[3], NULL, setObj,
			  TCL_PARSE_PART1|TCL_LEAVE_ERR_MSG)) {
	if (subObj) {
	  Tcl_DecrRefCount(subObj);
	}
	return TCL_ERROR;
      }
    }

    Tcl_SetObjResult(interp, Tcl_NewIntObj(numMatches));
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_NewSwitchObjCmd --
 *
 *	This object-based procedure is invoked to process the "nreswitch" Tcl
 *	command. See the user documentation for details on what it does.
 *
 * Results:
 *	A standard Tcl object result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static int
Tcl_NewSwitchObjCmd(dummy, interp, objc, objv)
    ClientData dummy;		/* Not used. */
    Tcl_Interp *interp;		/* Current interpreter. */
    int objc;			/* Number of arguments. */
    Tcl_Obj *CONST objv[];	/* Argument objects. */
{
#define EXACT	0
#define GLOB	1
#define REGEXP	2
    int switchObjc, index;
    Tcl_Obj *CONST *switchObjv;
    Tcl_Obj *patternObj, *bodyObj;
    char *string, *pattern, *body;
    int splitObjs, length, patternLen, i, code, mode, matched, bodyIdx;
    static char *switches[] =
	    {"-exact", "-glob", "-regexp", "--", (char *) NULL};

    switchObjc = objc-1;
    switchObjv = objv+1;
    mode = EXACT;
    
    string = Tcl_GetStringFromObj(switchObjv[0], &length);
    while ((switchObjc > 0) && (*string == '-')) {
	if (Tcl_GetIndexFromObj(interp, switchObjv[0], switches,
				"option", 0, &index) != TCL_OK) {
	    return TCL_ERROR;
	}
	switch (index) {
	    case 0:			/* -exact */
		mode = EXACT;
		break;
	    case 1:			/* -glob */
		mode = GLOB;
		break;
	    case 2:			/* -regexp */
		mode = REGEXP;
		break;
	    case 3:			/* -- */
		switchObjc--;
		switchObjv++;
		goto doneWithSwitches;
	}
	switchObjc--;
	switchObjv++;
	string = Tcl_GetStringFromObj(switchObjv[0], &length);
    }

    doneWithSwitches:
    if (switchObjc < 2) {
	Tcl_WrongNumArgs(interp, 1, objv,
		"?switches? string pattern body ... ?default body?");
	return TCL_ERROR;
    }
    
    string = Tcl_GetStringFromObj(switchObjv[0], &length);
    switchObjc--;
    switchObjv++;

    /*
     * If all of the pattern/command pairs are lumped into a single
     * argument, split them out again.
     */

    splitObjs = 0;
    if (switchObjc == 1) {
	code = Tcl_ListObjLength(interp, switchObjv[0], &switchObjc);
	if (code != TCL_OK) {
	    return code;
	}
	splitObjs = 1;
    }

    for (i = 0;  i < switchObjc;  i += 2) {
	if (i == (switchObjc-1)) {
	    Tcl_ResetResult(interp);
	    Tcl_AppendToObj(Tcl_GetObjResult(interp),
	            "extra switch pattern with no body", -1);
	    code = TCL_ERROR;
	    goto done;
	}

	/*
	 * See if the pattern matches the string.
	 */

	if (splitObjs) {
	    code = Tcl_ListObjIndex(interp, switchObjv[0], i, &patternObj);
	    if (code != TCL_OK) {
		return code;
	    }
	} else {
	    patternObj = switchObjv[i];
	}
	pattern = Tcl_GetStringFromObj(patternObj, &patternLen);

	matched = 0;
	if ((*pattern == 'd') && (i == switchObjc-2)
		&& (strcmp(pattern, "default") == 0)) {
	    matched = 1;
	} else {
	    switch (mode) {
		case EXACT:
		    matched = 0;
		    if (length == patternLen) {
		      matched = (memcmp(string, pattern, (size_t)length) == 0);
		    }
		    break;
		case GLOB:
		  /*
		   * WARNING: will not work with data containing NULLs.
		   */
		    matched = Tcl_StringMatch(string, pattern);
		    break;
		case REGEXP:
		    matched = Tcl_RegExpMatchObj(interp, string, length,
						 patternObj);
		    if (matched < 0) {
			code = TCL_ERROR;
			goto done;
		    }
		    break;
	    }
	}
	if (!matched) {
	    continue;
	}

	/*
	 * We've got a match. Find a body to execute, skipping bodies
	 * that are "-".
	 */

	for (bodyIdx = i+1;  ;  bodyIdx += 2) {
	    if (bodyIdx >= switchObjc) {
		Tcl_AppendStringsToObj(Tcl_GetObjResult(interp),
			"no body specified for pattern \"", pattern,
			"\"", (char *) NULL);
		code = TCL_ERROR;
		goto done;
	    }
	    
	    if (splitObjs) {
		code = Tcl_ListObjIndex(interp, switchObjv[0], bodyIdx,
		        &bodyObj);
		if (code != TCL_OK) {
		    return code;
		}
	    } else {
		bodyObj = switchObjv[bodyIdx];
	    }
	    /*
	     * THIS FAILS IF AN OBJECT'S STRING REP HAS A NULL.
	     */
	    body = Tcl_GetStringFromObj(bodyObj, &length);
	    if ((length != 1) || (body[0] != '-')) {
		break;
	    }
	}
	code = Tcl_EvalObj(interp, bodyObj);
	if (code == TCL_ERROR) {
	    char msg[100];
	    sprintf(msg, "\n    (\"%.50s\" arm line %d)", pattern,
		    interp->errorLine);
	    Tcl_AddObjErrorInfo(interp, msg, -1);
	}
	goto done;
    }

    /*
     * Nothing matched:  return nothing.
     */

    code = TCL_OK;

    done:
    return code;
#undef EXACT
#undef GLOB
#undef REGEXP
}


#if defined(_WIN32) || defined(__WIN32__)
__declspec(dllexport)
#endif
int Nre_Init(interp)
Tcl_Interp* interp;
{
  int code;
  code = Tcl_PkgProvide(interp, (char*)"Nre", (char*)VERSION);
  if (code != TCL_OK) {
    return code;
  }

  Tcl_RegisterObjType(&tclRegExpType);
  Tcl_CreateObjCommand(interp, (char*)"nrematch",
		       (Tcl_ObjCmdProc*)Tcl_NewRegexpObjCmd,
		       NULL, NULL);
  Tcl_CreateObjCommand(interp, (char*)"nresub",
		       (Tcl_ObjCmdProc*)Tcl_NewRegsubObjCmd,
		       NULL, NULL);
  Tcl_CreateObjCommand(interp, (char*)"nredump",
		       (Tcl_ObjCmdProc*)Tcl_NewRegDumpObjCmd,
		       NULL, NULL);
  Tcl_CreateObjCommand(interp, (char*)"nrelsearch",
		       (Tcl_ObjCmdProc*)Tcl_NewLsearchObjCmd,
		       NULL, NULL);
  Tcl_CreateObjCommand(interp, (char*)"nreswitch",
		       (Tcl_ObjCmdProc*)Tcl_NewSwitchObjCmd,
		       NULL, NULL);
  return TCL_OK;
}
